require 'tst/reporters'

module Tst
  VERSION = "0.0.3"

  # An error class that carries data about test failures.
  class Failure < StandardError
    attr_reader :message, :expected, :actual

    def initialize(message, expected, actual)
      @message = message
      @expected = expected
      @actual = actual
      super(message)
    end
  end

  # Assertions available in `tst` blocks.
  module Assertions

    # Fails if `value` is `false` or `nil`. Succeeds otherwise.
    def assert(value)
      return if value
      raise Failure.new("Failure: Truthiness", "not false or nil", value)
    end

    # Fails unless it's arguments are equal (with `==`).
    def assert_equal(expected, actual)
      return if expected == actual
      raise Failure.new("Equality Failure", expected, actual)
    end

    # Succeeds if it catches an error AND that error is
    # a `kind_of?` the `expected` error. Fails otherwise.
    def assert_raises(expected=StandardError)
      begin
        yield
      rescue => actual
        return actual if actual.kind_of?(expected)
        raise Failure.new("Failure: Unexpected Exception", expected, actual)
      end
      raise Failure.new("Failure: No Exception", expected, "Nothing raised.")
    end
  end

  # Set up some constants to represent test outcomes.
  SUCCEEDED = :success
  FAILED = :failure
  RAISED = :exception

  # A little value object containing test result data
  Result = Struct.new(:name, :status, :elapsed, :exception)

  # A collection class that keeps track of test results.
  class Results
    attr_reader :count, :elapsed, :results

    def initialize
      @count = 0
      @elapsed = 0
      @results = []
    end

    # Collects a test result.
    def <<(result)
      @count += 1
      @elapsed += result.elapsed
      results << result
    end

    # Accessors to pull out the subset of tests you need.
    def successes; results.select { |r| r.status == SUCCEEDED } end
    def failures; results.select { |r| r.status == FAILED } end
    def exceptions; results.select { |r| r.status == RAISED } end
  end

  # Runs test files in a fresh scope. It's not perfectly isolated, since
  # requires can still add to the global scope, but it's good enough.
  class Runner

    # Gets us access to the assertions in our `tst` blocks.
    include Assertions

    def initialize(file, reporter, results)
      @__reporter = reporter
      @__results = results
      __run_tests(file)
    end

    # Runs the tests by `instance-eval`-ing the test file.
    def __run_tests(file)
      instance_eval(File.read(file), File.expand_path(file), 1)
    end

    # The `tst` methods itself.
    # Takes a `name` and a block.
    # Runs the block.
    # Records the result.
    def tst(name, &block)
      start = Time.now
      block.call
      status = SUCCEEDED
    rescue Failure => exception
      status = FAILED
    rescue StandardError => exception
      status = RAISED
    ensure
      elapsed = Time.now - start
      __record(name, status, elapsed, exception)
    end

    def __record(name, status, elapsed, exception=nil)
      result = Result.new(name, status, elapsed, exception)
      @__results << result
      @__reporter.send(status, result)
    end
  end

  # The main entry point.
  #
  # glob - a String defining which test files to run.
  #        Passed through to Dir[].
  #
  # Accepted options:
  #   :load_paths - an Array of directories to add to the load path.
  #   :reporter   - a reporter instance (see 'lib/tst/reporters.rb'
  #                 for docs on reporters)
  def self.run(glob, options={})
    options = default_options.merge(options)
    $:.unshift(*options[:load_paths])
    results = Results.new
    reporter = options[:reporter]
    Dir[glob].each { |file| Runner.new(file, reporter, results) }
    reporter.summarize(results)
  end

  # The default options for `run`.
  def self.default_options
    { :reporter => Reporters::Plain.new, :load_paths => [] }
  end
end
