module Tst
  module Reporters
    class Plain
      attr_reader :io, :results

      def initialize(io = $stdout)
        @io = io
      end

      def success(result)   io.print('.') end
      def failure(result)   io.print('F') end
      def exception(result) io.print('E') end

      def summarize(results)
        @results = results
        summary = ["\n"] + failures + exceptions + conclusion
        summary.each { |line| io.puts(line) }
        nil
      end

      def failures
        results.failures.each_with_index.flat_map do |failure, i|
          failure_details(failure.name, failure.exception, i)
        end
      end

      def failure_details(name, exception, i)
        [ failure_header(name, exception, i),
          expected_actual(exception),
          backtrace(exception.backtrace),
          "" ]
      end

      def failure_header(name, exception, i)
        "F#{i+1}: #{exception.message} - #{name.inspect}"
      end

      def expected_actual(exception)
        ["  exp: #{exception.expected.inspect}",
         "  got: #{exception.actual.inspect}"
        ].join("\n")
      end

      def exceptions
        results.exceptions.each_with_index.flat_map do |exception, i|
          exception_details(exception.name, exception.exception, i)
        end
      end

      def exception_details(name, exception, i)
        [ exception_header(name, exception, i),
          backtrace(exception.backtrace),
          "" ]
      end

      def exception_header(name, exception, i)
        "E#{i+1}: #{exception.inspect} - #{name.inspect}"
      end


      def backtrace(trace)
        filter_backtrace(trace).map { |line| "* #{line}" }.join("\n")
      end

      def conclusion
        [ "#{passed}, #{failed}, #{raised} in %.6fs" % results.elapsed ]
      end

      def passed; "#{results.successes.count} passed"  end
      def failed; "#{results.failures.count} failed"   end
      def raised; "#{results.exceptions.count} raised" end

      def filter_backtrace(backtrace)
        backtrace.reject do |line|
          line =~ BACKTRACE_FILTER
        end.map do |line|
          line.gsub("#{Dir.pwd}/", "").gsub(/:in.*$/, '')
        end
      end

      BACKTRACE_FILTER =
        /lib\/tst\.rb|bin\/tst|\/(ruby|jruby|rbx)[-\/]([0-9\.])+|Rakefile/
    end
  end
end
